package it.gotoandplay.smartfoxserver.data.buddylist;

import it.gotoandplay.smartfoxserver.SmartFoxServer;
import it.gotoandplay.smartfoxserver.data.User;
import it.gotoandplay.smartfoxserver.data.Zone;
import it.gotoandplay.smartfoxserver.data.buddylist.persistence.IBuddyListPersister;
import it.gotoandplay.smartfoxserver.exceptions.BuddyListException;
import it.gotoandplay.smartfoxserver.extensions.ExtensionHelper;
import it.gotoandplay.smartfoxserver.lib.IService;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Logger;

public class BuddyListManager
  implements IService
{
  public static final int SECURITY_MODE_BASIC = 0;
  public static final int SECURITY_MODE_ADVANCED = 1;
  public static final String DEFAULT_PERSISTER_CLASS = "it.gotoandplay.smartfoxserver.data.buddylist.persistence.DefaultBuddyListPersister";
  public static final String DEFAULT_ADVANCED_PERSISTER_CLASS = "it.gotoandplay.smartfoxserver.data.buddylist.persistence.AdvancedBuddyListPersister";
  public static final int DEFAULT_SIZE = 50;
  public static final int DEFAULT_MAX_BUDDY_VARS = 10;
  public static final String DEFAULT_USE_OFFLINE_VARIABLES = "false";
  public static final String DEFAULT_SECURITY_MODE = "basic";
  public static String OFFLINE_VARS_PREFIX = "$";
  public static final String DEFAULT_MUTUAL_SETTINGS = "true";
  public static final String DEFAULT_USE_PERMISSION = "true";
  public static final String DEFAULT_PERMISSION_TIMEOUT = "60";
  private Map<String, StorableBuddyList> buddyMap;
  private int maxLength;
  private int maxBuddyVars;
  private boolean useOfflineVariables;
  private boolean mutualAddBuddy;
  private boolean mutualRemoveBuddy;
  private boolean addBuddyPermission;
  private int buddyPermissionTimeout;
  private int securityMode;
  private String zoneName;
  private Zone zone;
  private IBuddyListPersister persister;
  private IOfflineVariableManager offlineVarCache;
  private Map<PermissionRequestKey, PermissionRequest> permissionRequests;

  public BuddyListManager(String zoneName, int maxLength)
  {
    this.mutualAddBuddy = true;
    this.mutualRemoveBuddy = true;
    this.addBuddyPermission = true;
    this.buddyPermissionTimeout = 60;

    this.buddyMap = new ConcurrentHashMap();
    this.zoneName = zoneName;
    this.maxLength = maxLength;
    this.zone = SmartFoxServer.getInstance().getZone(this.zoneName);
    this.maxBuddyVars = 10;

    this.offlineVarCache = new OfflineBuddyVariablesManager();
    this.permissionRequests = new ConcurrentHashMap();

    this.useOfflineVariables = false;
    this.securityMode = 0;
  }

  public BuddyListManager(String zoneName, int maxLength, IBuddyListPersister persister)
  {
    this(zoneName, maxLength);
    this.persister = persister;
  }

  public void init(Object param)
  {
  }

  public void destroy(Object param)
  {
    this.persister.saveAll(this.buddyMap);
    this.persister.destroy(null);
  }

  public IBuddyListPersister getPersister()
  {
    return this.persister;
  }

  public void setPersister(IBuddyListPersister persister)
  {
    this.persister = persister;
  }

  public void addBuddy(String owner, String buddyName)
    throws BuddyListException
  {
    boolean isNew = false;

    StorableBuddyList bList = (StorableBuddyList)this.buddyMap.get(owner);

    if (bList == null)
    {
      bList = new StorableBuddyList(owner);
      isNew = true;
    }

    if (bList.buddies.size() >= this.maxLength)
    {
      throw new BuddyListException("The buddylist for user: " + owner + " is full");
    }

    synchronized (bList.buddies)
    {
      StorableBuddyItem item = new StorableBuddyItem(buddyName, false);

      if (!(bList.buddies.contains(item))) {
        bList.buddies.add(item);
      }
      else
        throw new BuddyListException("Duplicate buddy item. Owner: " + owner + ", Buddy: " + buddyName);

    }

    if (isNew)
      this.buddyMap.put(owner, bList);
  }

  public boolean removeBuddy(String owner, String buddyName)
  {
    boolean success = false;

    StorableBuddyList bList = (StorableBuddyList)this.buddyMap.get(owner);

    if (bList != null)
    {
      synchronized (bList.buddies)
      {
        success = bList.buddies.remove(new StorableBuddyItem(buddyName));
      }

      if (this.securityMode == 1)
      {
        boolean bool1;
        bList = (StorableBuddyList)this.buddyMap.get(buddyName);
        boolean bool1 = false;

        if (bList != null)
        {
          synchronized (bList.buddies)
          {
            bool1 = bList.buddies.remove(new StorableBuddyItem(owner));
          }

        }

        this.persister.removeOfflineBuddy(buddyName, owner);

        if ((success) && (bool1))
          SmartFoxServer.log.fine("Mutual Remove Buddy: " + owner + ", " + buddyName);
      }
    }

    return success;
  }

  public StorableBuddyList getBuddyList(String name)
  {
    return ((StorableBuddyList)this.buddyMap.get(name));
  }

  public boolean saveBuddyList(String userName)
  {
    boolean success = true;
    StorableBuddyList buddyList = (StorableBuddyList)this.buddyMap.get(userName);

    if (buddyList != null)
      success = this.persister.saveList(userName, buddyList);

    return success;
  }

  public void removeBuddyList(String owner)
  {
    this.buddyMap.remove(owner);
  }

  public void saveAndRemove(String owner)
  {
    saveBuddyList(owner);
    removeBuddyList(owner);
  }

  public void saveAllBuddyLists()
  {
    this.persister.saveAll(this.buddyMap);
  }

  public LoadableBuddyList loadBuddyList(String userName)
  {
    LoadableBuddyList buddyList = new LoadableBuddyList();

    StorableBuddyList storedList = (StorableBuddyList)this.buddyMap.get(userName);

    if (storedList == null)
    {
      storedList = this.persister.loadList(userName);

      if (storedList == null)
        storedList = new StorableBuddyList(userName);

      this.buddyMap.put(userName, storedList);
    }

    buddyList.variables = storedList.variables;

    List bList = storedList.buddies;

    if ((bList != null) && (bList.size() > 0))
    {
      for (Iterator it = bList.iterator(); it.hasNext(); )
      {
        StorableBuddyItem buddyItem = (StorableBuddyItem)it.next();
        int roomId = -1;
        boolean status = this.zone.checkBuddy(buddyItem.name);
        Integer uid = this.zone.getUserIdByName(buddyItem.name);

        int id = (uid != null) ? uid.intValue() : -1;

        LoadableBuddyItem loadableBuddyItem = new LoadableBuddyItem(buddyItem.name, 
          status, 
          roomId, 
          id, 
          buddyItem.isBlocked);

        loadableBuddyItem.setVariables(buddyItem.variables);

        buddyList.buddies.add(loadableBuddyItem);
      }

    }

    this.offlineVarCache.removeVariables(userName);

    return buddyList;
  }

  public List<String> setBuddyVariables(User user, Map<String, String> variables)
  {
    String userName = user.getName();
    StorableBuddyList storedList = (StorableBuddyList)this.buddyMap.get(userName);

    if (storedList == null)
    {
      storedList = new StorableBuddyList(userName);
      this.buddyMap.put(userName, storedList);
    }

    Iterator iter = variables.entrySet().iterator();

    List changedVars = new ArrayList();

    while (iter.hasNext())
    {
      Map.Entry item = (Map.Entry)iter.next();
      String vName = (String)item.getKey();
      String vValue = (String)item.getValue();

      SmartFoxServer.log.finer("Setting buddyVariable: " + vName + " = " + vValue);

      if ((this.maxBuddyVars > 0) && (user.buddyVariables.size() <= this.maxBuddyVars))
      {
        String oldValue = (String)user.buddyVariables.get(vName);

        if (oldValue != vValue)
        {
          user.buddyVariables.put(vName, vValue);
          changedVars.add(vName);
        }

      }
      else
      {
        SmartFoxServer.log.info("Too many buddy variables for user: " + user.getName());
        break;
      }
    }

    return changedVars;
  }

  public LinkedList getClientsToUpdate(String buddyName)
  {
    LinkedList channelsToUpdate = new LinkedList();
    Set keyList = null;
    StorableBuddyList bList = null;
    String owner = null;

    keyList = this.buddyMap.keySet();
    StorableBuddyItem tempBuddy = new StorableBuddyItem(buddyName);

    for (Iterator it = keyList.iterator(); it.hasNext(); )
    {
      owner = (String)it.next();

      bList = (StorableBuddyList)this.buddyMap.get(owner);

      if (bList.buddies.size() <= 0)
        break label136;
      label136: synchronized (bList.buddies)
      {
        if (bList.buddies.contains(tempBuddy))
        {
          channelsToUpdate.add(this.zone.getChannelFromName(owner));
        }
      }

    }

    return channelsToUpdate;
  }

  public void clearExpiredPermissionRequests()
  {
    long now = System.currentTimeMillis();

    for (Iterator it = this.permissionRequests.values().iterator(); it.hasNext(); )
    {
      PermissionRequest request = (PermissionRequest)it.next();

      if (now <= request.creationTime + this.buddyPermissionTimeout * 1000)
        break label83;
      SmartFoxServer.log.finer("Removing expired buddy list permission request: " + request);
      label83: it.remove();
    }
  }

  public void clearBuddyList(String owner)
  {
    synchronized (this.buddyMap)
    {
      StorableBuddyList bList = (StorableBuddyList)this.buddyMap.get(owner);

      if (bList != null)
        bList.buddies.clear();
    }
  }

  public int getMaxBuddyVars()
  {
    return this.maxBuddyVars;
  }

  public void setMaxBuddyVars(int maxBuddyVars)
  {
    this.maxBuddyVars = maxBuddyVars;
  }

  public boolean isBuddyOnline(String userName)
  {
    return this.zone.checkBuddy(userName);
  }

  public Map<String, String> getBuddyVariables(String buddyName)
  {
    Map variables = null;

    User user = this.zone.getUserByName(buddyName);

    if (user != null)
      variables = user.buddyVariables;

    return variables;
  }

  public Map<String, String> getBuddyVariables(LoadableBuddyItem buddy)
  {
    return getBuddyVariables(buddy.getName());
  }

  public void setBuddyVariables(String ownerName, Map<String, String> variables)
  {
    User user = this.zone.getUserByName(ownerName);

    if (user != null)
      user.buddyVariables = variables;
  }

  public Map<String, String> getOfflineBuddyVariables(String ownerName)
  {
    if (!(this.useOfflineVariables)) {
      return null;
    }

    Map vars = this.offlineVarCache.getVariables(ownerName);

    if (vars == null)
    {
      vars = this.persister.getOfflineVariables(ownerName);
      this.offlineVarCache.setVariables(ownerName, vars);
    }

    return vars;
  }

  public void sendBuddyListUpdate(User user)
  {
    ExtensionHelper.instance().sendBuddyListUpdate(user);
  }

  public boolean getUseOfflineVariables()
  {
    return this.useOfflineVariables;
  }

  public void setUseOfflineVariables(boolean useOfflineVariables)
  {
    this.useOfflineVariables = useOfflineVariables;
  }

  public int getSecurityMode()
  {
    return this.securityMode;
  }

  public void setSecurityMode(int securityMode)
  {
    this.securityMode = securityMode;
  }

  public boolean isAdvancedMode()
  {
    return (this.securityMode == 1);
  }

  public Map<PermissionRequestKey, PermissionRequest> getPermissionRequests()
  {
    return this.permissionRequests;
  }

  public boolean isMutualAddBuddy()
  {
    return ((this.mutualAddBuddy) && (this.securityMode == 1));
  }

  public void setMutualAddBuddy(boolean mutualAddBuddy)
  {
    this.mutualAddBuddy = mutualAddBuddy;
  }

  public boolean isMutualRemoveBuddy()
  {
    return ((this.mutualRemoveBuddy) && (this.securityMode == 1));
  }

  public void setMutualRemoveBuddy(boolean mutualRemoveBuddy)
  {
    this.mutualRemoveBuddy = mutualRemoveBuddy;
  }

  public int getBuddyPermissionTimeout()
  {
    return this.buddyPermissionTimeout;
  }

  public void setBuddyPermissionTimeout(int buddyPermissionTimeout)
  {
    this.buddyPermissionTimeout = buddyPermissionTimeout;
  }

  public void setAddBuddyPermission(boolean addBuddyPermission)
  {
    this.addBuddyPermission = addBuddyPermission;
  }

  public boolean isAddBuddyPermission()
  {
    return this.addBuddyPermission;
  }

  public void setBuddyBlockStatus(User owner, String buddyName, boolean status)
  {
    if ((buddyName != null) && (!(buddyName.equals(""))))
    {
      StorableBuddyList bList = getBuddyList(owner.getName());

      synchronized (bList.buddies)
      {
        for (Iterator it = bList.buddies.iterator(); it.hasNext(); )
        {
          StorableBuddyItem item = (StorableBuddyItem)it.next();

          if (!(item.name.equals(buddyName)))
            break label85;
          item.isBlocked = status;
          owner.setBuddyListChanged();

          label85: break;
        }
      }
    }
  }

  public boolean validateRecipient(User sender, User recipient)
  {
    boolean valid = false;

    if ((sender != null) && (recipient != null))
    {
      StorableBuddyItem buddyItem = getBuddyFromList(sender, recipient.getName());
      boolean recipientIsNotBlockedInSenderBuddyList = true;

      if (buddyItem != null)
        recipientIsNotBlockedInSenderBuddyList = !(buddyItem.isBlocked);

      buddyItem = getBuddyFromList(recipient, sender.getName());
      boolean senderIsNotBlockedInRecipientBuddyList = true;

      if (buddyItem != null)
        senderIsNotBlockedInRecipientBuddyList = !(buddyItem.isBlocked);

      boolean mutualBuddiesActiveAndBothBuddiesValid = (isMutualAddBuddy()) && (recipientIsNotBlockedInSenderBuddyList) && (senderIsNotBlockedInRecipientBuddyList);
      boolean nonMutualAndBuddyNotBlocked = (!(isMutualAddBuddy())) && (senderIsNotBlockedInRecipientBuddyList);

      if ((mutualBuddiesActiveAndBothBuddiesValid) || (nonMutualAndBuddyNotBlocked))
        valid = true;
    }

    return valid;
  }

  private StorableBuddyItem getBuddyFromList(User owner, String buddyName)
  {
    StorableBuddyItem buddItem = null;

    StorableBuddyList bList = getBuddyList(owner.getName());

    if (bList != null)
    {
      synchronized (bList.buddies)
      {
        for (Iterator it = bList.buddies.iterator(); it.hasNext(); )
        {
          StorableBuddyItem item = (StorableBuddyItem)it.next();

          if (!(item.name.equals(buddyName)))
            break label71;
          buddItem = item;
          label71: break;
        }
      }

    }

    return buddItem;
  }

  public void checkValidAddRequest(User sender, String target)
    throws BuddyListException
  {
    PermissionRequestKey key = new PermissionRequestKey(sender.getName(), target);
    PermissionRequest request = (PermissionRequest)this.permissionRequests.get(key);

    if (request != null)
      throw new BuddyListException("An identical request already exist!");

    if (this.zone.getUserByName(target) == null)
      throw new BuddyListException("Target user is off-line");

    if (getBuddyFromList(sender, target) != null)
      throw new BuddyListException("Buddy already exists in requester's buddy list!");
  }

  public void addPermissionRequest(String sender, String target)
  {
    PermissionRequestKey key = new PermissionRequestKey(sender, target);
    PermissionRequest request = new PermissionRequest(sender, target, System.currentTimeMillis());

    this.permissionRequests.put(key, request);
  }

  public boolean setPermissionStatus(String sender, String target, PermissionStatus status)
  {
    boolean result = false;

    PermissionRequestKey key = new PermissionRequestKey(sender, target);
    PermissionRequest request = (PermissionRequest)this.permissionRequests.get(key);

    if ((request != null) && (request.status == PermissionStatus.PENDING))
    {
      request.status = status;
      result = request.status == PermissionStatus.GRANTED;
    }
    else {
      request.status = PermissionStatus.REJECTED;
    }
    return result;
  }

  public Map<String, StorableBuddyList> getBuddyLists()
  {
    return this.buddyMap;
  }

  public Map<String, Map<String, String>> getVarCache()
  {
    return ((OfflineBuddyVariablesManager)this.offlineVarCache).getCache();
  }

  public Map<PermissionRequestKey, PermissionRequest> getPendingRequests()
  {
    return this.permissionRequests;
  }
}