package it.gotoandplay.smartfoxserver.extensions;

import it.gotoandplay.smartfoxserver.SmartFoxServer;
import it.gotoandplay.smartfoxserver.controllers.ExtensionHandler;
import it.gotoandplay.smartfoxserver.lib.ActionscriptObject;
import java.io.File;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.logging.Logger;

public class ExtensionManager
{
  private Map extensions;
  private boolean isShuttingDown;
  private Map extensionLastSave;

  public ExtensionManager()
  {
    this.extensions = new HashMap();
    this.isShuttingDown = false;
    this.extensionLastSave = new HashMap();
  }

  public void add(String key, ISmartFoxExtension ext)
  {
    if (!(this.isShuttingDown))
    {
      synchronized (this.extensions)
      {
        this.extensions.put(key, ext);
      }
    }
  }

  public boolean contains(String key)
  {
    synchronized (this.extensions)
    {
      return this.extensions.containsKey(key);
    }
  }

  private void remove(String key)
  {
    if (!(this.isShuttingDown))
    {
      synchronized (this.extensions)
      {
        this.extensions.remove(key);
      }
    }
  }

  public ISmartFoxExtension get(String key)
  {
    ISmartFoxExtension sfe;
    if (!(this.isShuttingDown))
      sfe = (ISmartFoxExtension)this.extensions.get(key);
    else
      sfe = null;

    return sfe;
  }

  public int getSize()
  {
    return this.extensions.size();
  }

  public void shutDown()
  {
    if (!(this.isShuttingDown))
    {
      synchronized (this.extensions)
      {
        this.isShuttingDown = true;

        for (Iterator i = this.extensions.entrySet().iterator(); i.hasNext(); )
        {
          Map.Entry item = (Map.Entry)i.next();

          SmartFoxServer.log.fine("Shutting down extension: " + item.getKey().toString());

          ISmartFoxExtension sfe = (ISmartFoxExtension)item.getValue();

          sfe.unRegister();

          sfe.destroy();
        }
      }
    }
  }

  public void activateAll()
  {
    setAllExtensionsActivityStatus(true);
  }

  public void deactivateAll()
  {
    setAllExtensionsActivityStatus(false);
  }

  private void setAllExtensionsActivityStatus(boolean b)
  {
    synchronized (this.extensions)
    {
      for (Iterator i = this.extensions.values().iterator(); i.hasNext(); )
      {
        ISmartFoxExtension sfe = (ISmartFoxExtension)i.next();

        if (sfe == null) break label48;
        label48: sfe.setActive(b);
      }
    }
  }

  public void toggleStatus(String xtName)
  {
    ISmartFoxExtension ext = get(xtName);

    if (ext != null)
    {
      ext.setActive(!(ext.isActive()));
      SmartFoxServer.log.info("Extension [ " + xtName + " ] was " + ((ext.isActive()) ? "activated" : "deactivated"));
    }
  }

  public void reloadExtension(String name)
  {
    ISmartFoxExtension extension = get(name);
    ISmartFoxExtension newExt = null;

    if (extension != null)
    {
      int extType;
      String extSourceFile;
      String zOwner = ((AbstractExtension)extension).getOwnerZone();
      String rOwner = ((AbstractExtension)extension).getOwnerRoom();

      if (extension instanceof JavascriptExtension)
      {
        extType = 2;
        extSourceFile = ((JavascriptExtension)extension).getScriptFileName();
      }
      else if (extension instanceof PythonExtension)
      {
        extType = 3;
        extSourceFile = ((PythonExtension)extension).getScriptFileName();
      }
      else
      {
        extType = 1;
        extSourceFile = extension.getClass().getName();
      }

      newExt = SmartFoxServer.getInstance().getExtensionHandler().loadExtension(name, extSourceFile, extType);

      if (newExt != null)
      {
        extension.unRegister();

        extension.destroy();

        remove(name);

        newExt.registerForEvents(zOwner, rOwner);

        newExt.init();

        add(name, newExt);

        StringBuffer sb = new StringBuffer("Extension [ ");
        sb.append(name).append(" ] reloaded.");

        SmartFoxServer.log.info(sb.toString());
      }
      else
      {
        SmartFoxServer.log.warning("Failed reloading extension: " + name);
      }
    } else {
      SmartFoxServer.log.warning("Invalid Extension reloading request: " + name);
    }
  }

  public void automagicReload()
  {
    synchronized (this.extensions)
    {
      Set exts = this.extensions.entrySet();
      ArrayList toReload = new ArrayList();

      for (Iterator i = exts.iterator(); i.hasNext(); )
      {
        int extType;
        String extSourceFile;
        Map.Entry item = (Map.Entry)i.next();
        String key = (String)item.getKey();
        ISmartFoxExtension ext = (ISmartFoxExtension)item.getValue();

        if (ext instanceof JavascriptExtension)
        {
          extType = 2;
          extSourceFile = "./sfsExtensions/" + ((JavascriptExtension)ext).getScriptFileName();
        }
        else if (ext instanceof PythonExtension)
        {
          extType = 3;
          extSourceFile = "./sfsExtensions/" + ((PythonExtension)ext).getScriptFileName();
        }
        else
        {
          extType = 1;
          Class clazz = ext.getClass();

          extSourceFile = clazz.getResource(clazz.getSimpleName() + ".class").toString().substring(5);
        }

        if (this.extensionLastSave.containsKey(extSourceFile))
        {
          long lastSaveTime = ((Long)this.extensionLastSave.get(extSourceFile)).longValue();

          File f = new File(extSourceFile);
          long currSaveTime = f.lastModified();

          if (currSaveTime <= lastSaveTime)
            break label334;
          this.extensionLastSave.remove(extSourceFile);
          this.extensionLastSave.put(extSourceFile, new Long(currSaveTime));

          toReload.add(key); break label334:
        }

        File f = new File(extSourceFile);
        label334: this.extensionLastSave.put(extSourceFile, new Long(f.lastModified()));
      }

      if (toReload.size() > 0)
      {
        for (i = toReload.iterator(); i.hasNext(); )
        {
          reloadExtension((String)i.next());
        }
      }
    }
  }

  public void getExtensionList(ActionscriptObject ao)
  {
    int type = 0;

    int count = 0;

    synchronized (this.extensions)
    {
      if (this.extensions.size() > 0)
      {
        for (Iterator i = this.extensions.entrySet().iterator(); i.hasNext(); )
        {
          String extSrc;
          Map.Entry item = (Map.Entry)i.next();
          Object o = item.getValue();

          String className = o.getClass().toString();
          if (className.indexOf("JavascriptExtension") != -1)
          {
            type = 1;
            extSrc = ((JavascriptExtension)item.getValue()).getScriptFileName();
          }
          else if (className.indexOf("PythonExtension") != -1)
          {
            type = 2;
            extSrc = ((PythonExtension)item.getValue()).getScriptFileName();
          }
          else
          {
            extSrc = className.substring(className.lastIndexOf(".") + 1);
          }

          String extName = (String)item.getKey();

          ActionscriptObject aoItem = new ActionscriptObject();

          AbstractExtension ext = (AbstractExtension)o;

          aoItem.put("name", extName);
          aoItem.put("src", extSrc);
          aoItem.put("type", String.valueOf(type));
          aoItem.put("status", String.valueOf(ext.isActive()));

          ao.put(count++, aoItem);
        }
      }
    }
  }
}